package org.expeditee.auth.mail.gui;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Base64;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;

import javax.crypto.SecretKey;

import org.expeditee.auth.AuthenticatorBrowser;
import org.expeditee.auth.mail.Mail;
import org.expeditee.auth.mail.Mail.MailEntry;
import org.expeditee.core.Clip;
import org.expeditee.core.Colour;
import org.expeditee.core.Dimension;
import org.expeditee.core.Font;
import org.expeditee.core.Image;
import org.expeditee.encryption.Actions;
import org.expeditee.gio.EcosystemManager;
import org.expeditee.gio.GraphicsManager;
import org.expeditee.gui.DisplayController;
import org.expeditee.gui.FrameCreator;
import org.expeditee.gui.FrameGraphics;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.MessageBay;
import org.expeditee.items.Item;
import org.expeditee.items.Text;

public class MailBay {
	// Frameset name for mail
	public static final String EXPEDITEE_MAIL_FRAMESET_NAME = "expediteemail";
	
	// Font used for messages in the MailBay
	private static final Font MESSAGE_FONT = new Font("Serif-Plain-16");
	
	// Bay preview area constants.
	private static final int SPACING = 25;
	private static final int OFFSET_Y = 15;
	private static final int OFFSET_X = 20;
	
	private static final int MAIL_LINK_X = 50;

	private static List<Item> previewMessages = new ArrayList<Item>();
	private static FrameCreator creator;
	private static Mail mail;
	private static Text mailLink;
	private static Text checkMailAction;
	
	private static Image mailBufferImage;
	
	public static Mail getMailClient() {
		return mail;
	}
	
	public static Mail getMailClient(String user) {
		boolean restore = mail != null;
		String oldUser = restore ? mail.getUser() : null;
		MailBay.reconnectToUser(user);
		Mail ret = mail;
		if (restore) {
			MailBay.reconnectToUser(oldUser);
		}
		return ret;
	}

	public static List<Item> getPreviewMessages() { 
		return previewMessages;
	}
	
	public static boolean isPreviewMailItem(Item i) { 
		return previewMessages.contains(i) || i == getMailLink() || i == checkMailAction;
	}
	
	public static Image getImage(Clip clip, Dimension size) { 
		// Can't get an image with an invalid size
		if (size == null || size.width <= 0 || size.height <= 0) {
			return null;
		}
		
		// Update the buffer
		updateBuffer(Item.DEFAULT_BACKGROUND, clip, size);
		
		// Return the image buffer
		return mailBufferImage;
	}
	
	public static Item getMailLink() { 
		return mailLink;
	}
	
	public static Item getCheckMailAction() {
		return checkMailAction;
	}
	
	public static void reconnectToUser(String user) {
		mail = new Mail(user);
		creator = new FrameCreator(EXPEDITEE_MAIL_FRAMESET_NAME, FrameIO.MAIL_PATH, 
				EXPEDITEE_MAIL_FRAMESET_NAME, FrameCreator.ExistingFramesetOptions.AppendAfterLastItem,
				false, AuthenticatorBrowser.PROFILEENCRYPTIONLABEL);
		updateLink();
	}
	
	public static void checkMail() {
		// Invalidate whole area
		DisplayController.invalidateArea(DisplayController.getMessageBayPaintArea());
		previewMessages.clear();
		
		// Get new mail since last time we checked for mail
		List<MailEntry> newMail = mail.checkMail();
		
		// Place up to two previews
		Map<String, Integer> countBySender = countBySender(newMail);
		String[] senders = countBySender.keySet().toArray(new String[] {});	
		for (int i = 0; i < senders.length && i < 3; i++) {
			String sender = senders[i];
			Text text = new Text("You have received " + countBySender.get(sender) + " new messages from " + sender);
			text.setFont(MESSAGE_FONT);
			text.setPosition(MAIL_LINK_X, OFFSET_Y + (SPACING * (i + 1)));
			previewMessages.add(text);
		}
		
		// Add new mail to creator
		for (MailEntry mail: newMail) {
			String timestamp = mail.getTimestamp();
			String sender = mail.getSender();
			generateMailHeader(timestamp, sender);
			creator.addText("Subject: " + mail.getSubject(), Colour.BLACK, null, null, false).setFont(MESSAGE_FONT);
			creator.addText(mail.getMessage(), Colour.BLACK, null, null, false).setFont(MESSAGE_FONT);
			Map<String, String> optionsTextActionMap = mail.getOptionsTextActionMap();
			String[] optionsTextActionMapKeyset = optionsTextActionMap.keySet().toArray(new String[] {});
			for (int i = 0; i < optionsTextActionMapKeyset.length; i++) {
				String text = optionsTextActionMapKeyset[i];
				String[] split = text.split(":::");
				String action = optionsTextActionMap.get(text);
				Text t = new Text(split[0]);
				for (int o = 1; o < split.length; o++) {
					t.addToData(split[o]);
				}
				int y = OFFSET_Y + (previewMessages.size() + i) * SPACING;
				t.setPosition(OFFSET_X, y);
				t.setColor(Colour.BLUE);
				t.setFont(MESSAGE_FONT);
				t.setAction(action);
				creator.addItem(t, false);
			}
		}

		DisplayController.requestRefresh(true);
	}
		
	public static void decryptOneOffSecureMessage(SecretKey key, List<String> data) {
		byte[] subjectBytes = Base64.getDecoder().decode(data.get(3));
		String subject = new String(Actions.DecryptSymmetric(subjectBytes, key));
		byte[] messageBytes = Base64.getDecoder().decode(data.get(4));
		String message = new String(Actions.DecryptSymmetric(messageBytes, key));
		
		creator.addText(data.get(0) + " (Single-use encrypted message)", Colour.BLACK, null, null, false);
		creator.addText(subject, Colour.BLACK, null, null, false).setFont(MESSAGE_FONT);
		creator.addText(message, Colour.BLACK, null, null, false).setFont(MESSAGE_FONT);
		
		for (int i = 5; i < data.size(); i += 2) {
			byte[] optionKeyBytes = Base64.getDecoder().decode(data.get(i));
			String k = new String(org.expeditee.encryption.Actions.DecryptSymmetric(optionKeyBytes, key));
			byte[] optionValueBytes = Base64.getDecoder().decode(data.get(i + 1));
			String v = new String(org.expeditee.encryption.Actions.DecryptSymmetric(optionValueBytes, key));
			creator.addText(k, Colour.BLACK, null, v, false).setFont(MESSAGE_FONT);
		}
	}

	private static void generateMailHeader(String timestamp, String sender) {
		Text header = creator.addText(timestamp + " (From: " + sender + ")", Colour.BLACK, null, null, false);
		Path credentialsFilePath = Paths.get(FrameIO.CONTACTS_PATH).resolve(sender + "-credentials").resolve("credentials.inf");
		if (credentialsFilePath.toFile().exists()) {
			try (Scanner in = new Scanner(credentialsFilePath)) {
				String line = in.nextLine();
				int number = Integer.parseInt(line.replace(".exp", ""));
				header.setLink(sender + "-credentials" + number); 
			} catch (IOException e) {
				MessageBay.errorMessage("Unable to find credentials frame of message sender: " + sender);
			}
		}
	}
	
	private synchronized static void updateBuffer(Colour background, Clip clip, Dimension size) {
		// If the buffer doesn't exist or is the wrong size, recreate it
		if (mailBufferImage == null || !mailBufferImage.getSize().equals(size)) {
			mailBufferImage = Image.createImage(size, true);
			clip = null; // Need to recreate the entire image;
			updateSize();
		}
		
		// Prepare graphics
		GraphicsManager g = EcosystemManager.getGraphicsManager();
		g.pushDrawingSurface(mailBufferImage);
		if (clip != null) {
			g.pushClip(clip);
		}
		g.clear(background);
		g.setFont(MESSAGE_FONT);
		
		// Paint header
		FrameGraphics.PaintItem(getHeader());
		
		// Paint the mail messages to the bay
		for (Item message: previewMessages) {
			if (message == null) { continue; }
			
			if (clip == null || clip.isNotClipped() || message.isInDrawingArea(clip.getBounds())) {
				FrameGraphics.PaintItem(message);
			}
		}
		
		// Paint the status from the MessageBay
		Item status = MessageBay.getStatus();
		if (status != null) {
			FrameGraphics.PaintItem(status);
		}
		
		// Paint the link to the mail frame
		if (clip == null || clip.isNotClipped() || mailLink.isInDrawingArea(clip.getBounds())) {
			FrameGraphics.PaintItem(mailLink);
		}
		
		// Paint the action item to check for new mail
		if (clip == null || clip.isNotClipped() || checkMailAction.isInDrawingArea(clip.getBounds())) {
			FrameGraphics.PaintItem(checkMailAction);
		}
		
		g.popDrawingSurface();
	}
	
	private static void updateSize() {
		for (Item i: previewMessages) {
			if (i != null) {
				i.setOffset(0, -DisplayController.getMessageBayPaintArea().getMinY());
			}
		}
		
		updateLink();
	}
	
	private static void updateLink() {
		mailLink = new Text(-2, "See Mail", Colour.BLACK, Colour.WHITE);
		mailLink.setFont(MESSAGE_FONT);
		mailLink.setOffset(0, -DisplayController.getMessageBayPaintArea().getMinY());
		mailLink.setAnchorRight(OFFSET_X);
		mailLink.setAnchorTop(OFFSET_Y);
		mailLink.setLink(creator.getCurrent());
		
		checkMailAction = new Text(-2, "Get New Mail", Colour.BLACK, Colour.WHITE);
		checkMailAction.setFont(MESSAGE_FONT);
		checkMailAction.setOffset(0, -DisplayController.getMessageBayPaintArea().getMinY());
		checkMailAction.setAnchorRight(OFFSET_X);
		checkMailAction.setAnchorTop(OFFSET_Y + SPACING);
		checkMailAction.setAction("CheckForNewMail");
	}

	private static Text getHeader() {
		Text header = new Text("You have [" + previewMessages.size() + "] new messages since last you check.");
		header.setPosition(OFFSET_X, OFFSET_Y);
		header.setOffset(0, -DisplayController.getFramePaintAreaHeight());
		header.setFont(MESSAGE_FONT);
		return header;
	}

	private static Map<String, Integer> countBySender(List<MailEntry> newMail) {
		Map<String, Integer> bySenderCount = new HashMap<String, Integer>();
		
		for (MailEntry mail: newMail) {
			String sender = mail.getSender();
			if (mail.isSingleUseEncryption()) {
				sender = "Encrypted Sender";
			}
			
			if (bySenderCount.containsKey(sender)) {
				Integer newValue = bySenderCount.get(sender) + 1;
				bySenderCount.put(sender, newValue);
			} else {
				bySenderCount.put(sender, 1);
			}
		}
		
		return bySenderCount;
	}
}
